@@ -1,5 +1,6 @@
-import  marshal, os, psutil, random, sys, threading, time
+import  os, psutil, random, sys, threading, time
 from subprocess import PIPE
+from anvil_downlink_util.pipes import MessagePipe
 
 # State representing which workers are running here
 from anvil_downlink_host import workers_by_id, send_with_header, maybe_quit_if_draining_and_done, \
@@ -13,6 +14,7 @@ CACHE_LOCK = threading.Lock()
 
 class Worker:
     def __init__(self, first_req_id, enable_profiling=False, app_id=None, app_version=None, cache_key=None, set_timeout=True, task_info=None):
+        self.initial_req_id = first_req_id
         self.req_ids = {first_req_id}
         self.outbound_ids = {} # Outbound ID -> inbound ID it came from
         self.lock = threading.RLock()
@@ -22,6 +24,8 @@ class Worker:
                                        bufsize=0, stdin=PIPE, stdout=PIPE, preexec_fn=get_demote_fn(app_id))
         self.proc_info = psutil.Process(self.proc.pid)
         self.task_info = task_info
+        self.from_worker = MessagePipe(self.proc.stdout)
+        self.to_worker = MessagePipe(self.proc.stdin)
 
         workers_by_id[first_req_id] = self
         self.cache_key = cache_key
@@ -45,6 +49,7 @@ class Worker:
         self.start_time = {first_req_id: time.time()}
         self.enable_profiling = {first_req_id: enable_profiling}
         self.killing_task = False
+        self.global_error = None
 
         threading.Thread(target=self.read_loop).start()
 
@@ -100,6 +105,10 @@ class Worker:
         self.timed_out = True
         self.proc.terminate()
 
+    def kill_with_error(self, err):
+        self.global_error = err
+        self.proc.terminate()
+
     def drain(self):
         # Finish up our execution and time out.
         if self.cache_key is None:
@@ -133,6 +142,7 @@ class Worker:
         if self.killing_task:
             return
         self.killing_task = True
+        print("SOFT KILL BACKGROUND TASK %s" % self.initial_req_id)
 
         # Request state. If it returns with in 5 seconds, we will die with state, else we hard-kill
         self.timeout_timer = threading.Timer(5, self._hard_kill_background_task)
@@ -140,10 +150,10 @@ class Worker:
         self.timeout_timer.start()
 
     def _hard_kill_background_task(self):
-        print("TIMEOUT KILLING BACKGROUND TASK %s" % self.req_ids)
+        print("HARD KILL BACKGROUND TASK %s" % self.initial_req_id)
         try:
             self.req_ids.discard('pre-kill-task-state')
-            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': list(self.req_ids)[0]})
+            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': self.initial_req_id})
         finally:
             self.hard_timeout()
 
@@ -151,11 +161,10 @@ class Worker:
     def read_loop(self):
         try:
             while True:
-                # marshal.load() takes the GIL, so only do it once we know there's something there to load.
-                dummy_char = self.proc.stdout.read(1)
-                if len(dummy_char) == 0:
+                try:
+                    msg, bindata = self.from_worker.receive()
+                except EOFError:
                     break
-                msg = marshal.load(self.proc.stdout)
                 type = msg.get("type")
                 id = msg.get("id") or msg.get("requestId")
 
@@ -177,13 +186,11 @@ class Worker:
                         print("Discarding invalid message with bogus ID: %s" % repr(msg))
                         if type == "CHUNK_HEADER":
                             print("Discarding binary data chunk")
-                            marshal.load(self.proc.stdout)
                         continue
 
                 try:
                     if type == "CHUNK_HEADER":
-                        x = marshal.load(self.proc.stdout)
-                        send_with_header(msg, x)
+                        send_with_header(msg, bindata)
                         if msg.get("lastChunk"):
                             self.transmitted_media(msg['requestId'], msg['mediaId'])
                     else:
@@ -218,9 +225,12 @@ class Worker:
 
                             self.req_ids.discard('pre-kill-task-state')
 
-                            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': list(self.req_ids)[0],
+                            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': self.initial_req_id,
                                                          'taskState': msg['response'], 'objects': objects})
 
+                            if self.timeout_timer:
+                                self.timeout_timer.cancel()
+
                             self.proc.terminate()
                         else:
                             send_with_header(msg)
@@ -229,6 +239,9 @@ class Worker:
                         #if statsd and (id in self.start_time):
                         #    statsd.timing('Downlink.WorkerLifetime', (time.time()*1000) - self.start_time.get(id, 0)*1000)
                         self.on_media_complete(msg, lambda: self.responded(id))
+                        if "error" in msg and msg.get("moduleLoadFailed"):
+                            self.kill_with_error(msg["error"])
+
 
                 except UnicodeError:
                     send_with_header({"id": id, "error": {"type": "UnicodeError", "message": "This function returned a binary string (not text). If you want to return binary data, use a BlobMedia object instead."}})
@@ -253,71 +266,67 @@ class Worker:
 
             error_id = "".join([random.choice('0123456789abcdef') for x in range(10)])
             for i in self.req_ids:
-                if self.timed_out:
-                    message = 'Server code took too long'
-                    type = "anvil.server.TimeoutError"
+                if self.global_error is not None:
+                    err = self.global_error
+                elif self.timed_out:
+                    err = {
+                        'message': "Server code took too long",
+                        'type': "anvil.server.TimeoutError"
+                    }
                 elif rt == -9:
-                    message = 'Server code execution process was killed. It may have run out of memory: %s' % (error_id)
-                    type = "anvil.server.ExecutionTerminatedError"
-                    sys.stderr.write(message + " (IDs %s)\n" % i)
+                    err = {
+                        'message': "Server code execution process was killed. It may have run out of memory: %s" % (error_id),
+                        'type': "anvil.server.ExecutionTerminatedError"
+                    }
+                    sys.stderr.write(err['message'] + " (IDs %s)\n" % i)
                     sys.stderr.flush()
                 else:
-                    message = 'Server code exited unexpectedly: %s' % (error_id)
-                    type = "anvil.server.ExecutionTerminatedError"
-                    sys.stderr.write(message + " (IDs %s)\n" % i)
+                    err = {
+                        'message': "Server code exited unexpectedly: %s" % (error_id),
+                        'type': "anvil.server.ExecutionTerminatedError"
+                    }
+                    sys.stderr.write(err['message'] + " (IDs %s)\n" % i)
                     sys.stderr.flush()
-                send_with_header({'id': i, 'error':{'type': type, 'message': message}})
+                send_with_header({'id': i, 'error': err})
             print ("Worker terminated for IDs %s (return code %s)" % (self.req_ids, rt))
             maybe_quit_if_draining_and_done()
 
-    def send(self, msg, bin=False):
+    def send(self, msg, bindata=None):
+        id = msg.get("id")
+        if msg.get("type") in ["CALL", "CALL_WITH_APP", "GET_TASK_STATE", "LAUNCH_REPL", "REPL_COMMAND", "TERMINATE_REPL"]:
+            # It's a new request! Start the timeout
+            #print ("Setting timeout and routing for new request ID %s" % id)
+            workers_by_id[id] = self
+            if msg.get("enable-profiling"):
+                self.enable_profiling[id] = True
+                self.start_time[id] = time.time()
+            self.req_ids.add(id)
+            if msg["type"] != "REPL_COMMAND":
+                self.set_timeout(id)
 
-        if not bin:
-            id = msg.get("id")
-            if msg.get("type") in ["CALL", "CALL_WITH_APP", "GET_TASK_STATE", "LAUNCH_REPL", "REPL_COMMAND", "TERMINATE_REPL"]:
-                # It's a new request! Start the timeout
-                #print ("Setting timeout and routing for new request ID %s" % id)
-                workers_by_id[id] = self
-                if msg.get("enable-profiling"):
-                    self.enable_profiling[id] = True
-                    self.start_time[id] = time.time()
-                self.req_ids.add(id)
-                if msg["type"] != "REPL_COMMAND":
-                    self.set_timeout(id)
+        elif msg.get("type") == "REPL_KEEPALIVE":
+            self.clear_timeout(msg["repl"])
+            self.set_timeout(msg["repl"])
+            send_with_header({"id": id, "response": None})
+            return
 
-            elif msg.get("type") == "REPL_KEEPALIVE":
-                self.clear_timeout(msg["repl"])
-                self.set_timeout(msg["repl"])
-                send_with_header({"id": id, "response": None})
-                return
+        self.to_worker.send(msg, bindata)
 
-            # A horrid hack - a one-char "activation" that's not marshalled, because marshal holds the GIL and Windows doesn't support select() on pipes and urrrggghhh
+        def outbound_done():
+            self.outbound_ids.pop(id, None)
+            workers_by_id.pop(id, None)
+            maybe_quit_if_draining_and_done()
 
-            d = b"X" + marshal.dumps(msg)
-        else:
-            d = marshal.dumps(msg)
-
-        self.proc.stdin.write(d)
-        self.proc.stdin.flush()
-
-        if not bin:
-            def outbound_done():
-                self.outbound_ids.pop(id, None)
-                workers_by_id.pop(id, None)
-                maybe_quit_if_draining_and_done()
-
-            if "response" in msg or "error" in msg:
-                self.on_media_complete(msg, outbound_done)
+        if "response" in msg or "error" in msg:
+            self.on_media_complete(msg, outbound_done)
 
     def get_task_state(self, msg):
         self.send(msg)
 
-    def handle_inbound_message(self, msg, bin=None):
-        self.send(msg)
-        if bin is not None:
-            self.send(bin, bin=True)
-            if msg.get("last_chunk"):
-                self.transmitted_media(msg.get("requestId"), msg.get("mediaId"))
+    def handle_inbound_message(self, msg, bindata=None):
+        self.send(msg, bindata)
+        if bin is not None and msg.get("last_chunk"):
+            self.transmitted_media(msg.get("requestId"), msg.get("mediaId"))
 
     report_stats = report_worker_stats
 
